''' Audit utilities '''

def get_data_from_field(obj, field):
    ''' Takes an model obj and a field object, gets the value for the field
    from the supplied object and returns the resulting data
    '''
    field_type = field.get_internal_type()
    if field_type == 'ForeignKey':
        return getattr(obj, field.attname)
    elif field_type == 'DateTimeField':
        return str(getattr(obj, field.name))
    else:
        return getattr(obj, field.name)

def serialize_data(instance, relations):
    ''' Gets the data from every standard field and places it in the data
    dictionary. Then proceeds to inspect the relations list provided, and
    grabs all the necessary data by following those relations. When complete
    returns the data dictionary to be serialized
    '''
    data = {}
    for field in instance._meta.fields:
        data[field.name] = get_data_from_field(instance, field)

    for relation in relations:
        related_attr = getattr(instance, relation)
        data[relation] = []
        for item in related_attr.all():
            relation_data = {}
            for field in item._meta.fields:
                relation_data[field.name] = get_data_from_field(
                        item, field)
            data[relation].append(relation_data)

    return data

def data_has_changes(obj, relations, prev_audit=None):
    ''' Compares an obj to the previous audit item we created. If the
    serialized differs, return True. Otherwise, we return False.
    '''
    if not prev_audit:
        return True
    data = serialize_data(obj, relations)
    return data != prev_audit.audit_data



''' Tests for chronicler.models.create_audit '''
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User

from chronicler.models import AuditItem, create_audit
from chronicler.tests import TestCase
from chronicler.tests.models import Person, Group, Membership


class TestCreateAudit(TestCase):

    def test_create_audit(self):
        ''' Tests audit.models.create_audit receiver. Asserts that we properly
        create an AuditItem
        '''
        person = Person.objects.create(name='Tester')
        user, _ = User.objects.get_or_create(username='analyte')
        content_type = ContentType.objects.get_for_model(Person)
        group = Group.objects.create(name='group_one')
        group_two = Group.objects.create(name='group_two')
        member_one = Membership.objects.create(
            person=person,
            group=group,
            gold_member=True
        )
        member_two = Membership.objects.create(
            person=person,
            group=group_two,
            gold_member=False
        )

        create_audit(
            Person,
            person,
            ['membership_set'],
            user
        )
        audit = AuditItem.objects.get(
            content_type=content_type,
            object_id=person.pk
        )
        data = audit.audit_data
        audit_member_one, audit_member_two = data['membership_set']
        self.assertEqual(audit_member_one['id'], member_one.pk)
        self.assertEqual(audit_member_two['id'], member_two.pk)
        self.assertEqual(audit_member_one['group'], member_one.group.pk)
        self.assertEqual(audit_member_two['group'], member_two.group.pk)
        self.assertEqual(audit_member_one['gold_member'], True)
        self.assertEqual(audit_member_two['gold_member'], False)

        member_two.gold_member = True
        member_two.save()
        create_audit(
            Person,
            person,
            ['membership_set'],
            user
        )
        audit = AuditItem.objects.order_by('-pk')[0]
        self.assertEqual(
            audit.audit_data['membership_set'][1]['gold_member'],
            True
        )



''' Test models for running the chronicler test suite '''
from django.db import models


class Person(models.Model):

    name = models.CharField(max_length=12)


class Group(models.Model):

    name = models.CharField(max_length=12)
    people = models.ManyToManyField('Person', through='Membership')


class Membership(models.Model):

    person = models.ForeignKey('Person')
    group = models.ForeignKey('Group')
    gold_member = models.BooleanField(default=False)



''' Audit Log Models '''
import json

from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django import dispatch

from .utils import serialize_data, data_has_changes

audit = dispatch.Signal(
        providing_args=['instance', 'relations', 'user', 'force'])


class AuditItem(models.Model):

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')
    serialized_data = models.TextField(null=True, blank=True)
    user = models.ForeignKey('auth.User', null=True)
    created = models.DateTimeField(auto_now_add=True)

    @property
    def audit_data(self):
        return json.loads(self.serialized_data)

    class Meta:
        ordering = ('-created',)
        get_latest_by = 'created'


@dispatch.receiver(audit, weak=False, dispatch_uid='audit.create_audit')
def create_audit(sender, instance, relations, user, force=False, **kws):
    data = serialize_data(instance, relations)
    ct = ContentType.objects.get_for_model(instance)
    try:
        prev_audit = AuditItem.objects.filter(content_type=ct).latest()
    except AuditItem.DoesNotExist:
        prev_audit = None
    if data_has_changes(instance, relations, prev_audit) or force:
        AuditItem.objects.create(
            content_object=instance,
            user=user,
            serialized_data=json.dumps(data)
        )



from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.test import TransactionTestCase


class TestCase(TransactionTestCase):
    apps = ('chronicler.tests', )

    def _pre_setup(self):
        # Add the models to the db.
        self._original_installed_apps = list(settings.INSTALLED_APPS)
        settings.INSTALLED_APPS = list(settings.INSTALLED_APPS)
        for app in self.apps:
            settings.INSTALLED_APPS.append(app)
        apps.loaded = False
        call_command('migrate', 'tests',  '--noinput')
        # Call the original method that does the fixtures etc.
        super(TestCase, self)._pre_setup()

    def _post_teardown(self):
        # Call the original method.
        super(TestCase, self)._post_teardown()
        # Restore the settings.
        settings.INSTALLED_APPS = self._original_installed_apps
        apps.loaded = False


''' Decorators for chronicler '''
import functools

from django.utils.decorators import available_attrs

from .models import audit


def audits(model, relations, field_data_key, req_data_key, data_loc=None, force=False):
    ''' Creates an AuditItem with the data provided after a request is
    processed.

    Arguments:
        model: Model that we will be creating an audit entry from
        relations: Relations for the model object we need to track down
        field_data_key: The field on the model we should query against
        req_data_key: The key we should use to get identifying info from
        data_loc: GET, POST, or None. Specifies where we find our data to
        query with.
    '''

    def decorator(view_func):
        @functools.wraps(view_func, assigned=available_attrs(view_func))
        def _wrapped_view(request, *args, **kwargs):
            if data_loc == 'POST':
                value = request.POST.get(req_data_key)
            elif data_loc == 'GET':
                value = request.GET.get(req_data_key)
            else:
                value = kwargs.get(req_data_key)

            if not value:
                # Couldn't find requested req_data_key in POST/GET
                return view_func(request, *args, **kwargs)

            obj_kwargs = {field_data_key: value}
            obj = model.objects.get(**obj_kwargs)
            response = view_func(request, *args, **kwargs)
            audit.send(
                sender=model,
                instance=obj,
                relations=relations,
                user=request.user,
                force=force
            )
            return response
        return _wrapped_view
    return decorator



''' Tests for chronicler.decorators.audits '''
from mock import Mock

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType

from chronicler.models import AuditItem
from chronicler.decorators import audits

from chronicler.tests import TestCase
from chronicler.tests.models import Person, Group

@audits(Person, ['group_set'], 'pk', 'person_pk', 'POST')
def fake_view_post(request):
    pass


@audits(Person, ['group_set'], 'pk', 'person_pk', 'GET')
def fake_view_get(request):
    pass


@audits(Person, ['group_set'], 'pk', 'person_pk')
def fake_view(request, person_pk):
    pass


@audits(Person, ['group_set'], 'pk', 'person_pk', force=True)
def fake_view_force(request, person_pk):
    pass


class TestCreateAuditEntry(TestCase):

    def setUp(self):
        super(TestCreateAuditEntry, self).setUp()
        self.user, _ = User.objects.get_or_create(username='analyte')
        self.content_type = ContentType.objects.get_for_model(Person)
        self.person = Person.objects.create(name='Tester')

    def test_create_post_key(self):
        ''' Test that asserts we can get our object from the POST variables
        when necessary
        '''
        assert not AuditItem.objects.all()
        request = Mock(POST={'person_pk': self.person.pk}, user=self.user)
        fake_view_post(request)
        assert AuditItem.objects.filter(
            content_type=self.content_type,
            object_id=self.person.pk)

    def test_create_get_key(self):
        ''' Test that asserts we can get our object from the GET variables
        when necessary
        '''
        assert not AuditItem.objects.all()
        request = Mock(GET={'person_pk': self.person.pk}, user=self.user)
        fake_view_get(request)
        assert AuditItem.objects.filter(
            content_type=self.content_type,
            object_id=self.person.pk)

    def test_create_simple_view(self):
        ''' Test that proves that we can grab our necessary data to get an
        object from the request path
        '''
        assert not AuditItem.objects.all()
        request = Mock(user=self.user)
        fake_view(request, person_pk=self.person.pk)
        assert AuditItem.objects.filter(
            content_type=self.content_type,
            object_id=self.person.pk)

    def test_prevent_audit_dupes(self):
        ''' Test that asserts that when nothing changes, we don't create
        another audit item with identical changes
        '''
        assert not AuditItem.objects.all()
        request = Mock(user=self.user)
        fake_view(request, person_pk=self.person.pk)
        assert AuditItem.objects.filter(
            content_type=self.content_type,
            object_id=self.person.pk)

        fake_view(request, person_pk=self.person.pk)
        audit_items = AuditItem.objects.filter(
            content_type=self.content_type,
            object_id=self.person.pk)
        self.assertEqual(audit_items.count(), 1)

    def test_audits_force_create_dupes(self):
        ''' Test that asserts that even when we find nothing changes,
        that we will create a dupe if force is set to True
        '''
        assert not AuditItem.objects.all()
        request = Mock(user=self.user)
        fake_view(request, person_pk=self.person.pk)
        assert AuditItem.objects.filter(
            content_type=self.content_type,
            object_id=self.person.pk)

        fake_view_force(request, person_pk=self.person.pk)
        audit_items = AuditItem.objects.filter(
            content_type=self.content_type,
            object_id=self.person.pk)
        self.assertEqual(audit_items.count(), 2)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-20 21:45
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Group',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=12)),
            ],
        ),
        migrations.CreateModel(
            name='Membership',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('gold_member', models.BooleanField(default=False)),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Group')),
            ],
        ),
        migrations.CreateModel(
            name='Person',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=12)),
            ],
        ),
        migrations.AddField(
            model_name='membership',
            name='person',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Person'),
        ),
        migrations.AddField(
            model_name='group',
            name='people',
            field=models.ManyToManyField(through='tests.Membership', to='tests.Person'),
        ),
    ]



#!/usr/bin/env python

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages


install_requires = [
]

setup(
    name='chronicler',
    version='0.2-beta',
    author='Analyte Health',
    author_email='open-source@analytehealth.com',
    url='http://github.com/analytehealth/chronicler',
    description='A Django App for preserving revisions of modified Model Objects',
    packages=find_packages(),
    zip_safe=False,
    install_requires=install_requires,
    include_package_data=True,
    entry_points={},
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
)
